package zinger.io.file;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;

public class Console extends Frame implements FileListener
{
    protected class PreferencesDialog extends Dialog
    {
        protected TextField delayField;
        protected TextField fontSizeField;
        protected TextField maxCharsDisplayedField;
        protected Button cancelButton;
        protected Button okayButton;

        protected long delay;
        protected Font font;
        protected int maxCharsDisplayed;

        protected Boolean accepted = null;

        public PreferencesDialog()
        {
            super(Console.this, "Preferences", true);
            this.setResizable(false);
        }

        protected void windowInit()
        {
            this.fieldsInit();
            this.buttonsInit();

            Panel centerPanel = new Panel();
            centerPanel.setLayout(new GridLayout(3, 2));
            centerPanel.add(new Label("Delay"));
            centerPanel.add(this.delayField);
            centerPanel.add(new Label("Font Size"));
            centerPanel.add(this.fontSizeField);
            centerPanel.add(new Label("Max Chars Back"));
            centerPanel.add(this.maxCharsDisplayedField);

            Panel southPanel = new Panel();
            southPanel.setLayout(new FlowLayout());
            southPanel.add(this.okayButton);
            southPanel.add(this.cancelButton);

            this.setLayout(new BorderLayout());
            this.add(southPanel, BorderLayout.SOUTH);
            this.add(centerPanel, BorderLayout.CENTER);

            this.pack();
        }

        protected void fieldsInit()
        {
            this.delayField = new TextField();
            this.fontSizeField = new TextField();
            this.maxCharsDisplayedField = new TextField();
        }

        protected void buttonsInit()
        {
            this.cancelButton = new Button("Cancel");
            this.cancelButton.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent event)
                {
                    Console.PreferencesDialog.this.cancel();
                }
            });

            this.okayButton = new Button("Okay");
            this.okayButton.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent event)
                {
                    Console.PreferencesDialog.this.accept();
                }
            });
        }

        public void cancel()
        {
            this.accepted = Boolean.FALSE;
            this.hide();
        }

        public void accept()
        {
            try
            {
                this.delay = Long.parseLong(this.delayField.getText()) * 1000;
                this.font = this.font.deriveFont(Float.parseFloat(this.fontSizeField.getText()));
                this.maxCharsDisplayed = Integer.parseInt(this.maxCharsDisplayedField.getText());
                this.accepted = Boolean.TRUE;
            }
            catch(NumberFormatException ex)
            {
                this.accepted = Boolean.FALSE;
            }
            finally
            {
                this.hide();
            }
        }

        public void show()
        {
            if(this.accepted == null)
            {
                this.windowInit();
            }
            this.accepted = Boolean.FALSE;

            this.delayField.setText(String.valueOf(Console.FILE_WATCHER.getDelay() / 1000));
            this.font = Console.this.outputArea.getFont();
            this.fontSizeField.setText(String.valueOf(this.font.getSize()));
            this.maxCharsDisplayedField.setText(String.valueOf(Console.this.maxCharsDisplayed));

            super.show();
        }

        public boolean wasAccepted()
        {
            return this.accepted.booleanValue();
        }

        public long getDelay()
        {
            return this.delay;
        }

        public Font getFont()
        {
            return this.font;
        }

        public int getMaxCharsDisplayed()
        {
			return this.maxCharsDisplayed;
		}
    }

    protected static final String NO_FILE_TITLE = "File Watcher";

    protected static final FileWatcher FILE_WATCHER = new FileWatcher();

    private static int counter = 0;

    protected File file;

    protected TextArea outputArea;
    protected FileDialog fileDialog;
    protected PreferencesDialog preferencesDialog;

    private final StringBuffer stringBuffer = new StringBuffer();
    private final char[] charBuffer = new char[1 << 7];
    private long textLength = 0;
    private int maxCharsDisplayed = 2048;
    private boolean canAppend = false;

    public Console()
    {
        super(Console.NO_FILE_TITLE);

        this.frameInit();
    }

    protected void frameInit()
    {
        this.addWindowListener(new WindowAdapter()
        {
            public void windowClosing(WindowEvent ev)
            {
                if(ev.getWindow() == Console.this)
                {
                    Console.this.dispose();
                }
            }
        });

        this.fileDialogInit();

        this.preferencesDialogInit();

        this.outputAreaInit();

        Font font = new Font("Monospaced", 0, 14);
        this.outputArea.setFont(font);

        this.menuInit();
        this.add(outputArea);
    }

    protected void fileDialogInit()
    {
        this.fileDialog = new FileDialog(this);
    }

    protected void preferencesDialogInit()
    {
        this.preferencesDialog = this.new PreferencesDialog();
    }

    protected void outputAreaInit()
    {
        this.outputArea = new TextArea("", 0, 0, TextArea.SCROLLBARS_VERTICAL_ONLY);
        this.outputArea.setEditable(false);
    }

    protected void menuInit()
    {
        MenuItem openMenuItem = new MenuItem("Open");
        openMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                if(Console.this.file != null)
                {
                    Console.this.fileDialog.setDirectory(Console.this.file.getParent());
                }
                Console.this.fileDialog.show();
                Console.this.openFile(Console.this.fileDialog.getDirectory(), Console.this.fileDialog.getFile());
            }
        });

        MenuItem newWindowMenuItem = new MenuItem("New Window");
        newWindowMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                (new Console()).showFrame();
            }
        });

        MenuItem exitMenuItem = new MenuItem("Exit");
        exitMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                System.exit(0);
            }
        });

        Menu fileMenu = new Menu("File");
        fileMenu.add(openMenuItem);
        fileMenu.add(newWindowMenuItem);
        fileMenu.add(new MenuItem("-"));
        fileMenu.add(exitMenuItem);

        MenuItem copyMenuItem = new MenuItem("Copy");
        copyMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                Clipboard clipboard = Console.this.outputArea.getToolkit().getSystemClipboard();
                StringSelection selection = new StringSelection(Console.this.outputArea.getSelectedText());
                clipboard.setContents(selection, selection);
            }
        });

        MenuItem selectAllMenuItem = new MenuItem("Select All");
        selectAllMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                Console.this.outputArea.selectAll();
            }
        });

        Menu editMenu = new Menu("Edit");
        editMenu.add(copyMenuItem);
        editMenu.add(selectAllMenuItem);

        MenuItem clearMenuItem = new MenuItem("Clear");
        clearMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                Console.this.outputArea.setText("");
            }
        });

        MenuItem tailMenuItem = new MenuItem("Tail");
        tailMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                Console.this.refresh(true);
                Console.this.outputArea.setCaretPosition(Integer.MAX_VALUE);
            }
        });

        MenuItem refreshMenuItem = new MenuItem("Refresh");
        refreshMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                Console.this.refresh(false);
            }
        });

        final CheckboxMenuItem pauseMenuItem = new CheckboxMenuItem("Pause", false);
        pauseMenuItem.addItemListener(new ItemListener()
        {
            public void itemStateChanged(ItemEvent event)
            {
                if(pauseMenuItem.getState())
                {
                    Console.FILE_WATCHER.removeFileListener(Console.this);
                }
                else
                {
                    Console.FILE_WATCHER.addFileListener(Console.this);
                }
            }
        });

        MenuItem propertiesMenuItem = new MenuItem("Properties");
        propertiesMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                synchronized(Console.this.preferencesDialog)
                {
                    Console.this.preferencesDialog.show();
                    if(Console.this.preferencesDialog.wasAccepted())
                    {
                        Console.FILE_WATCHER.setDelay(Console.this.preferencesDialog.getDelay());
                        Console.this.outputArea.setFont(Console.this.preferencesDialog.getFont());
                        Console.this.maxCharsDisplayed = Console.this.preferencesDialog.getMaxCharsDisplayed();
                    }
                }
            }
        });

        Menu viewMenu = new Menu("View");
        viewMenu.add(clearMenuItem);
        viewMenu.add(tailMenuItem);
        viewMenu.add(refreshMenuItem);
        viewMenu.add(pauseMenuItem);
        viewMenu.add(new MenuItem("-"));
        viewMenu.add(propertiesMenuItem);

        MenuBar menuBar = new MenuBar();
        menuBar.add(fileMenu);
        menuBar.add(editMenu);
        menuBar.add(viewMenu);
        this.setMenuBar(menuBar);
    }

    public synchronized void hide()
    {
        if(this.isVisible())
        {
            --Console.counter;
            if(Console.counter == 0)
            {
                System.exit(0);
            }
            super.hide();
        }
    }

    public synchronized void show()
    {
        if(!this.isVisible())
        {
            ++Console.counter;
            super.show();
        }
    }

    public Dimension getPreferredSize()
    {
        return new Dimension(600, 400);
    }

    public void openFile(String directory, String file)
    {
        this.openFile(file == null ? null : new File(directory, file));
    }

    public synchronized void openFile(File file)
    {
        if(file != null)
        {
            if(this.file == null)
            {
                Console.FILE_WATCHER.addFileListener(this);
            }
            this.canAppend = false;
            this.file = file;
            Console.FILE_WATCHER.addFile(this.file);
            try
            {
                this.setTitle(file.getCanonicalPath());
            }
            catch(IOException ex)
            {
                ex.printStackTrace();
                this.setTitle(Console.NO_FILE_TITLE);
            }
        }
    }

    public synchronized void refresh(boolean append)
    {
        if(this.file != null && this.file.exists())
        {
            try
            {
				int toRemove = this.outputArea.getText().length() - this.maxCharsDisplayed;
				if(toRemove > 0)
				{
					this.outputArea.replaceRange("", 0, toRemove);
				}

                Reader reader = new FileReader(this.file);
                if(append)
                {
                    reader.skip(this.textLength);
                }
                else
                {
                    this.textLength = 0;
                }
                this.stringBuffer.setLength(0);
                for(int bytesRead = reader.read(this.charBuffer); bytesRead > 0; bytesRead = reader.read(this.charBuffer))
                {
                    this.textLength += bytesRead;
                    this.stringBuffer.append(charBuffer, 0, bytesRead);
                }
                reader.close();
                /*
                for(int i = this.stringBuffer.indexOf("\r"); i >= 0; i = this.stringBuffer.indexOf("\r"))
                {
                    this.stringBuffer.deleteCharAt(i);
                }
                */
                String text = this.stringBuffer.toString();
                if(append)
                {
                    //this.textLength += text.length();
                    this.outputArea.append(text);
                }
                else
                {
                    //this.textLength = text.length();
                    this.outputArea.setText(text);
                }
            }
            catch(FileNotFoundException ex)
            {
                ex.printStackTrace();
                this.outputArea.setText(ex.getMessage());
            }
            catch(IOException ex)
            {
                ex.printStackTrace();
                this.outputArea.setText(ex.getMessage());
            }
        }
        else
        {
            this.outputArea.setText("");
        }
    }

    public void handleFileEvent(FileEvent event)
    {
        if(this.file != null && this.file.equals(event.getFile()))
        {
            switch(event.getType())
            {
                case FileEvent.MODIFIED:
                break;

                case FileEvent.DELETED:
                this.canAppend = false;
                break;

                case FileEvent.CREATED:
                this.canAppend = false;
                break;
            }
            this.refresh(this.canAppend);
            this.canAppend = true;
        }
    }

    public void showFrame()
    {
        this.pack();
        this.show();
    }

    public static void main(String[] args)
    {
        String delayProperty = System.getProperty("delay");
        try
        {
            if(delayProperty != null)
            {
                Console.FILE_WATCHER.setDelay(Long.parseLong(delayProperty));
            }
        }
        catch(NumberFormatException ex)
        {
            System.err.println("Delay property not understood.");
        }
        Console console = null;
        for(int i = 0; i < args.length; i++)
        {
            console = new Console();
            console.openFile(new File(args[i]));
            console.showFrame();
        }
        if(console == null)
        {
            console = new Console();
            console.showFrame();
        }
    }
}